#include <Wire.h>
#include <HD44780_LCD_PCF8574.h>
#include "RTC.h"
#include <RTClib.h>

HD44780LCD lcd(2, 16, 0x27, &Wire);
RTCTime internalTime;
RTC_DS3231 ds3231;

void setup() {
  Serial.begin(115200);
  while (!Serial);
  delay(200);

  lcd.PCF8574_LCDInit(lcd.LCDCursorTypeOff);    // LCD初期化（カーソル非表示）
  lcd.PCF8574_LCDClearScreen();                 // LCD画面クリア
  lcd.PCF8574_LCDBackLightSet(true);            // バックライトON

  RTC.begin();                                  // 内蔵RTC初期化
  if (!ds3231.begin()) {                        // DS3231初期化・エラー時はLCDに表示
    lcd.PCF8574_LCDSendString("DS3231 Error!");
    while (1);
  }

  Serial.println("Enter date and time as: YYYYMMDD HHMMSS");
  Serial.println("Example: 20250503 023300");
}

void loop() {
  static unsigned long lastUpdate = 0;
  char buffer[17];

  // 1秒ごとにLCD表示を更新
  if (millis() - lastUpdate >= 1000) {
    lastUpdate = millis();

    // 内蔵RTCの現在時刻を取得
    RTC.getTime(internalTime);

    // DS3231の現在時刻を取得
    DateTime now = ds3231.now();

    // UNIXタイム（エポック秒）を計算
    uint32_t ds3231_epoch = now.unixtime();
    DateTime internalDT(
      internalTime.getYear(),
      static_cast<int>(internalTime.getMonth()) + 1, // Monthは0始まりなので+1
      internalTime.getDayOfMonth(),
      internalTime.getHour(),
      internalTime.getMinutes(),
      internalTime.getSeconds()
    );
    uint32_t internal_epoch = internalDT.unixtime();

    // 差分計算（内蔵RTCが進んでいれば正、遅れていれば負）
    int32_t diff_sec = (int32_t)(internal_epoch - ds3231_epoch);

    // --- 右詰め表示の詳しい解説 ---
    // 1. bufferに"R4 hh:mm:ss"（11文字）を作成（左側に表示する内容）
    snprintf(buffer, sizeof(buffer), "R4 %02d:%02d:%02d",
             internalTime.getHour(),
             internalTime.getMinutes(),
             internalTime.getSeconds());

    // 2. diffbufに"+0000"や"-0015"など、差分（5文字）を作成（右側に表示する内容）
    //    %+5ldで符号付き・5桁・右詰めゼロ埋めなし
    char diffbuf[7];
    snprintf(diffbuf, sizeof(diffbuf), "%+5ld", diff_sec); // 例: "  +12" "  -15"

    // 3. line1に時刻＋差分を連結して16桁にする
    //    - buffer（"R4 hh:mm:ss"）は11文字
    //    - diffbufは5文字
    //    - 11+5=16文字でLCD1行分ピッタリ
    //    - "%s%5s"で時刻の後ろに差分を右詰めで配置
    char line1[17];
    snprintf(line1, sizeof(line1), "%s%5s", buffer, diffbuf);

    // 4. LCDの1行目に出力
    //    例: "R4 12:34:56  -15"
    lcd.PCF8574_LCDGOTO(lcd.LCDLineNumberOne, 0);
    lcd.PCF8574_LCDSendString(line1);

    // 2行目：DS3231の時分秒（"DS hh:mm:ss"）
    snprintf(buffer, sizeof(buffer), "DS %02d:%02d:%02d",
             now.hour(), now.minute(), now.second());
    lcd.PCF8574_LCDGOTO(lcd.LCDLineNumberTwo, 0);
    lcd.PCF8574_LCDSendString(buffer);
  }

  // シリアル入力監視（随時受付）
  if (Serial.available()) {
    String input = Serial.readStringUntil('\n');
    input.trim();

    // 入力例: 20250503 023300（15文字、スペース区切り）
    if (input.length() == 15 && input.charAt(8) == ' ') {
      // 各項目を分割して整数変換
      String y = input.substring(0,4);
      String m = input.substring(4,6);
      String d = input.substring(6,8);
      String H = input.substring(9,11);
      String M = input.substring(11,13);
      String S = input.substring(13,15);

      int year  = y.toInt();
      int month = m.toInt();
      int day   = d.toInt();
      int hour  = H.toInt();
      int min   = M.toInt();
      int sec   = S.toInt();

      // 値チェック
      bool valid = true;
      String msg = "";
      if (year < 2000 || year > 2099) { valid = false; msg += "Year "; }
      if (month < 1 || month > 12)     { valid = false; msg += "Month "; }
      if (day < 1 || day > 31)         { valid = false; msg += "Day "; }
      if (hour < 0 || hour > 23)       { valid = false; msg += "Hour "; }
      if (min < 0 || min > 59)         { valid = false; msg += "Minute "; }
      if (sec < 0 || sec > 59)         { valid = false; msg += "Second "; }
      // ※簡易チェック。2月や月末日数、うるう年等は省略

      if (valid) {
        // 内蔵RTCの時刻を設定
        RTCTime newTime(
          day,
          static_cast<Month>(month-1), // Month列挙型は0始まり
          year,
          hour, min, sec,
          DayOfWeek::SATURDAY, // 曜日は自動計算していません
          SaveLight::SAVING_TIME_INACTIVE
        );
        RTC.setTime(newTime);

        // DS3231の時刻を設定
        ds3231.adjust(DateTime(year, month, day, hour, min, sec));

        // 設定日時をYYYY-MM-DD HH:MM:SS形式で同じ行に表示
        char setbuf[24];
        snprintf(setbuf, sizeof(setbuf), "%04d-%02d-%02d %02d:%02d:%02d",
                 year, month, day, hour, min, sec);
        Serial.print("Time set! ");
        Serial.println(setbuf);

        lcd.PCF8574_LCDGOTO(lcd.LCDLineNumberOne, 0);
        lcd.PCF8574_LCDSendString("Set OK           ");
      } else {
        // 入力値が不正な場合のエラー表示
        Serial.print("Invalid: "); Serial.println(msg);
        lcd.PCF8574_LCDGOTO(lcd.LCDLineNumberOne, 0);
        lcd.PCF8574_LCDSendString("Invalid input    ");
      }
    } else if (input.length() > 0) {
      // フォーマットエラー時の表示
      Serial.println("Format error. Use: YYYYMMDD HHMMSS");
      lcd.PCF8574_LCDGOTO(lcd.LCDLineNumberOne, 0);
      lcd.PCF8574_LCDSendString("Format error     ");
    }
  }
}
